from asyncio import get_event_loop, set_event_loop, SelectorEventLoop
from functools import wraps
from operator import attrgetter, itemgetter

import requests
from flask import redirect, render_template, request, url_for, abort, json
from requests import RequestException

from . import utils
from .config import app, auth, logger, scheduler, ADMIN_USERNAME, ADMIN_PASSWORD, MYSQL_ENABLED, APP_PORT, IS_DEBUG, \
    CLOUDPROXY_ENDPOINT, TRANSMISSION_ENABLED
from .connectors import get_instance, run_all, Nyaa
from .connectors.core import ConnectorLang, ConnectorReturn
from .forms import SearchForm, DeleteForm, EditForm, FolderDeleteForm, FolderEditForm

if MYSQL_ENABLED:
    from .config import db
    from .models import AnimeFolder, AnimeTitle, AnimeLink

if TRANSMISSION_ENABLED:
    from .config import transmission


def mysql_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if not MYSQL_ENABLED:
            return abort(404)
        return f(*args, **kwargs)

    return decorated_function


def clean_titles():
    db.engine.execute("""
DELETE
FROM anime_title
WHERE id IN (
    SELECT anime_title.id
    FROM anime_title
    LEFT JOIN anime_link ON anime_title.id = anime_link.title_id
    WHERE anime_link.id IS NULL
)
""")


@auth.verify_password
def verify_password(username, password):
    return username == ADMIN_USERNAME and ADMIN_PASSWORD == password


@app.template_filter('boldify')
def boldify(name):
    query = request.args.get('q', '')
    name = utils.boldify(name, query)
    if MYSQL_ENABLED:
        for keyword in db.session.query(AnimeTitle.keyword.distinct()).all():
            if keyword[0].lower() != query.lower():
                name = utils.boldify(name, keyword[0])
    return name


@app.template_filter('flagify')
def flagify(is_vf):
    return ConnectorLang.FR.value if is_vf else ConnectorLang.JP.value


@app.template_filter('colorify')
def colorify(model):
    return get_instance(model.link, model.title.keyword).color


@app.context_processor
def inject_user():
    return dict(mysql_disabled=not MYSQL_ENABLED)


@app.route('/')
def home():
    return render_template('layout.html', search_form=SearchForm(), title='Anime torrents search engine')


@app.route('/search')
def search():
    query = request.args.get('q')
    if not query:
        return redirect(url_for('home'))

    set_event_loop(SelectorEventLoop())
    torrents = get_event_loop().run_until_complete(run_all(query))
    return render_template('search.html', search_form=SearchForm(), connectors=torrents)


@app.route('/latest')
@app.route('/latest/<int:page>')
def latest(page=1):
    set_event_loop(SelectorEventLoop())
    torrents = get_event_loop().run_until_complete(
        run_all('', return_type=ConnectorReturn.HISTORY, page=page)
    )

    results = []
    for torrent in torrents:
        results = results + torrent.data
    for result in results:
        result['self'] = get_instance(result['href'])
    results.sort(key=itemgetter('date'), reverse=True)

    return render_template('latest.html', search_form=SearchForm(), torrents=results, page=page)


@app.route('/list')
@app.route('/list/<url_filters>')
@mysql_required
def list_animes(url_filters='nyaa,yggtorrent'):
    filters = None
    for i, to_filter in enumerate(url_filters.split(',')):
        if not i:
            filters = AnimeLink.link.contains(to_filter)
        else:
            filters = filters | AnimeLink.link.contains(to_filter)

    titles = db.session.query(AnimeTitle, AnimeLink).join(AnimeLink).filter(filters).order_by(AnimeTitle.name).all()

    results = {}
    for title, link in titles:
        if title.id not in results:
            results[title.id] = [link]
        else:
            results[title.id].append(link)

    return render_template('list.html', search_form=SearchForm(), titles=results)


@app.route('/admin', methods=['GET', 'POST'])
@mysql_required
@auth.login_required
def admin():
    form = DeleteForm(request.form)

    if form.validate_on_submit():
        link = AnimeLink.query.filter_by(id=form.id.data).first()
        if link:
            form.message = '%s (%s) has been successfully deleted' % (
                link.title.name,
                link.season
            )
            db.session.delete(link)
            db.session.commit()

            title = link.title
            if title and not len(title.links):
                db.session.delete(title)
                db.session.commit()
        else:
            form._errors = {
                'id': ['Id %s was not found in the database' % form.id.data]
            }

    folders = AnimeFolder.query.all()
    for folder in folders:
        for title in folder.titles:
            title.links.sort(key=attrgetter('season'))
        folder.titles.sort(key=attrgetter('name'))

    return render_template('admin/list.html', search_form=SearchForm(), folders=folders, action_form=form)


@app.route('/admin/folder', methods=['GET', 'POST'])
@mysql_required
@auth.login_required
def folder_list():
    form = FolderDeleteForm(request.form)

    if form.validate_on_submit():
        folder = AnimeFolder.query.filter_by(id=form.id.data).first()
        if folder:
            form.message = '%s has been successfully deleted' % folder.name
            db.session.delete(folder)
            db.session.commit()
        else:
            form._errors = {
                'id': ['Id %s was not found in the database' % form.id.data]
            }

    folders = AnimeFolder.query.all()

    return render_template('admin/folder/list.html', search_form=SearchForm(), folders=folders, action_form=form)


@app.route('/admin/folder/edit', methods=['GET', 'POST'])
@app.route('/admin/folder/edit/<int:folder_id>', methods=['GET', 'POST'])
@mysql_required
@auth.login_required
def folder_edit(folder_id=None):
    folder = AnimeFolder.query.filter_by(id=folder_id).first()
    folder = folder if folder else AnimeFolder()
    form = FolderEditForm(
        request.form,
        id=folder.id,
        name=folder.name,
        path=folder.path
    )

    if form.validate_on_submit():
        # Folder
        folder.name = form.name.data
        folder.path = form.path.data
        db.session.add(folder)
        db.session.commit()
        return redirect(url_for('folder_list'))

    return render_template('admin/folder/edit.html', search_form=SearchForm(), action_form=form)


@app.route('/admin/edit', methods=['GET', 'POST'])
@app.route('/admin/edit/<int:link_id>', methods=['GET', 'POST'])
@mysql_required
@auth.login_required
def admin_edit(link_id=None):
    link = AnimeLink.query.filter_by(id=link_id).first()
    link = link if link else AnimeLink()

    folders = AnimeFolder.query.all()
    form = EditForm(
        request.form,
        id=link.id,
        folder=link.title.folder.id if link.title else None,
        name=link.title.name if link.title else None,
        link=link.link,
        season=link.season,
        comment=link.comment,
        keyword=link.title.keyword if link.title else None
    )
    form.folder.choices = [('', '')] + [(g.id, g.name) for g in folders]

    if form.validate_on_submit():
        # Instance for VF tag
        instance = get_instance(form.link.data)

        # Title
        title = AnimeTitle.query.filter_by(id=link.title_id).first()
        title = title if title else AnimeTitle.query.filter_by(name=form.name.data).first()
        title = title if title else AnimeTitle()
        title.folder_id = form.folder.data
        title.name = form.name.data
        title.keyword = form.keyword.data.lower()
        db.session.add(title)
        db.session.commit()

        # Link
        link.title_id = title.id
        link.link = form.link.data
        link.season = form.season.data
        link.comment = form.comment.data
        link.vf = instance.is_vf(form.link.data)

        # Database
        db.session.add(link)
        db.session.commit()
        clean_titles()

        # Transmission
        if TRANSMISSION_ENABLED and isinstance(instance, Nyaa):
            if title.folder.path is not None and title.folder.path != '':
                download_url = link.link.replace(
                    '/view/',
                    '/download/'
                ) + '.torrent'
                torrent_path = '%s/%s' % (title.folder.path, title.name)
                torrent = transmission.add_torrent(
                    download_url,
                    download_dir=torrent_path
                )
                transmission.move_torrent_data(torrent.id, torrent_path)
                transmission.start_torrent(torrent.id)

        return redirect(url_for('admin'))

    return render_template('admin/edit.html', search_form=SearchForm(), folders=folders, action_form=form)


@scheduler.task('interval', id='flaredestroyy', days=1)
def flaredestroyy():
    if CLOUDPROXY_ENDPOINT:
        try:
            json_session = requests.post(CLOUDPROXY_ENDPOINT, json={
                'cmd': 'sessions.list'
            })
            response = json.loads(json_session.text)
            sessions = response['sessions']

            for session in sessions:
                requests.post(CLOUDPROXY_ENDPOINT, json={
                    'cmd': 'sessions.destroy',
                    'session': session
                })
                logger.info('Destroyed %s' % session)
        except RequestException as e:
            logger.exception(e)


def run():
    scheduler.start()
    flaredestroyy()
    app.run('0.0.0.0', APP_PORT, IS_DEBUG)



from datetime import datetime, timedelta

from bs4 import BeautifulSoup

from .core import ConnectorCore, ConnectorReturn, ConnectorCache, curl_content
from ..utils import parse_date, link_exist_in_db


class AnimeUltime(ConnectorCore):
    color = 'is-warning'
    title = 'Anime-Ultime'
    favicon = 'animeultime.png'
    base_url = 'http://www.anime-ultime.net'
    is_light = True
    is_behind_cloudflare = False

    def get_full_search_url(self):
        from_date = ''
        sort_type = 'search'

        if self.return_type is ConnectorReturn.HISTORY:
            try:
                page_date = datetime.now() - timedelta((int(self.page) - 1) * 365 / 12)
            except OverflowError:
                page_date = datetime.fromtimestamp(0)
            from_date = page_date.strftime('%m%Y')
            sort_type = 'history'

        return '%s/%s-0-1/%s' % (self.base_url, sort_type, from_date)

    @ConnectorCache.cache_data
    def search(self):
        response = curl_content(self.get_full_search_url(), {
            'search': self.query
        })

        if response['http_code'] == 200:
            html = BeautifulSoup(response['output'], 'html.parser')
            title = html.select('div.title')
            player = html.select('div.AUVideoPlayer')

            if len(title) > 0 and 'Recherche' in title[0].get_text():
                trs = html.select('table.jtable tr')

                for i, tr in enumerate(trs):
                    if not i:
                        continue

                    tds = tr.findAll('td')

                    if len(tds) < 2:
                        continue

                    url = tds[0].a
                    href = '%s/%s' % (self.base_url, url['href'])

                    if not any(href == d['href'] for d in self.data):
                        self.data.append({
                            'vf': self.is_vf(),
                            'href': href,
                            'name': url.get_text(),
                            'type': tds[1].get_text(),
                            'class': self.color if link_exist_in_db(href) else ''
                        })
            elif len(player) > 0:
                name = html.select('h1')
                ani_type = html.select('div.titre')
                href = '%s/file-0-1/%s' % (
                    self.base_url,
                    player[0]['data-serie']
                )

                self.data.append({
                    'vf': self.is_vf(),
                    'href': href,
                    'name': name[0].get_text(),
                    'type': ani_type[0].get_text().replace(':', ''),
                    'class': self.color if link_exist_in_db(href) else ''
                })

            self.on_error = False

    @ConnectorCache.cache_data
    def get_history(self):
        response = curl_content(self.get_full_search_url())

        if response['http_code'] == 200:
            html = BeautifulSoup(response['output'], 'html.parser')
            tables = html.select('table.jtable')
            h3s = html.findAll('h3')

            for i, table in enumerate(tables):
                for j, tr in enumerate(table.findAll('tr')):
                    if not j:
                        continue

                    tds = tr.findAll('td')
                    link = tds[0].a
                    href = '%s/%s' % (self.base_url, link['href'])

                    self.data.append({
                        'vf': self.is_vf(),
                        'href': href,
                        'name': link.get_text(),
                        'type': tds[4].get_text(),
                        'date': parse_date(h3s[i].string[:-3], '%A %d %B %Y'),
                        'class': self.color if link_exist_in_db(href) else ''
                    })

            self.on_error = False

    @ConnectorCache.cache_data
    def is_vf(self, url=''):
        return False



import logging
from os import environ, urandom

from flask import Flask
from flask.cli import load_dotenv
from flask_apscheduler import APScheduler
from flask_httpauth import HTTPBasicAuth
from flask_sqlalchemy import SQLAlchemy
from redis import Redis
from transmission_rpc.client import Client

load_dotenv()

IS_DEBUG = environ.get('FLASK_ENV', 'production') == 'development'
ADMIN_USERNAME = environ.get('ADMIN_USERNAME', 'admin')
ADMIN_PASSWORD = environ.get('ADMIN_PASSWORD', 'secret')
APP_PORT = int(environ.get('FLASK_PORT', 5000))
CACHE_TIMEOUT = int(environ.get('CACHE_TIMEOUT', 60 * 60))
REQUESTS_TIMEOUT = int(environ.get('REQUESTS_TIMEOUT', 5))
BLACKLIST_WORDS = environ.get('BLACKLIST_WORDS', '').split(',') if environ.get('BLACKLIST_WORDS', '') else []
CLOUDPROXY_ENDPOINT = environ.get('CLOUDPROXY_ENDPOINT')
MYSQL_ENABLED = False
REDIS_ENABLED = False
TRANSMISSION_ENABLED = False

app = Flask(__name__)
app.name = 'PyNyaaTa'
app.debug = IS_DEBUG
app.secret_key = urandom(24).hex()
app.url_map.strict_slashes = False
auth = HTTPBasicAuth()
scheduler = APScheduler(app=app)
logging.basicConfig(level=(logging.DEBUG if IS_DEBUG else logging.INFO))
logger = logging.getLogger(app.name)

db_host = environ.get('MYSQL_SERVER')
if db_host:
    MYSQL_ENABLED = True
    db_user = environ.get('MYSQL_USER')
    db_password = environ.get('MYSQL_PASSWORD')
    db_name = environ.get('MYSQL_DATABASE')
    if not db_user or not db_password or not db_name:
        logger.error('Missing connection environment variables')
        exit()
    app.config['SQLALCHEMY_DATABASE_URI'] = 'mysql+pymysql://%s:%s@%s/%s?charset=utf8mb4' % (
        db_user, db_password, db_host, db_name
    )
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
    app.config['SQLALCHEMY_ECHO'] = IS_DEBUG
    app.config['SQLALCHEMY_ENGINE_OPTIONS'] = {
        'pool_recycle': 200
    }
    db = SQLAlchemy(app)
    from .models import create_all

    create_all()

cache_host = environ.get('REDIS_SERVER')
if cache_host:
    REDIS_ENABLED = True
    cache = Redis(cache_host)

transmission_host = environ.get('TRANSMISSION_SERVER')
if transmission_host:
    TRANSMISSION_ENABLED = True
    transmission_username = environ.get('TRANSMISSION_RPC_USERNAME')
    transmission_password = environ.get('TRANSMISSION_RPC_PASSWORD')
    transmission = Client(
        username=transmission_username,
        password=transmission_password,
        host=transmission_host,
        logger=logger
    )



from .config import db


class AnimeFolder(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(
        length=100,
        collation='utf8mb4_general_ci'
    ), unique=True, nullable=False)
    path = db.Column(db.String(length=100, collation='utf8mb4_general_ci'))
    titles = db.relationship(
        "AnimeTitle",
        backref="folder",
        cascade='all,delete-orphan'
    )


class AnimeTitle(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(
        length=100,
        collation='utf8mb4_general_ci'
    ), unique=True, nullable=False)
    keyword = db.Column(db.Text(
        collation='utf8mb4_general_ci'
    ), nullable=False)
    folder_id = db.Column(db.Integer, db.ForeignKey('anime_folder.id'))
    links = db.relationship(
        'AnimeLink',
        backref="title",
        cascade='all,delete-orphan'
    )


class AnimeLink(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    link = db.Column(db.Text(collation='utf8mb4_general_ci'), nullable=False)
    season = db.Column(db.Text(collation='utf8mb4_general_ci'), nullable=False)
    comment = db.Column(db.Text(collation='utf8mb4_general_ci'))
    vf = db.Column(db.Boolean, nullable=False)
    title_id = db.Column(db.Integer, db.ForeignKey('anime_title.id'))


def create_all():
    db.create_all()
